@@ -28,8 +28,9 @@ import sys
 import socket
 import decimal
 import datetime
+import time
 
-VERSION = (0, 2, 3)
+VERSION = (0, 2, 4)
 __version__ = '%s.%s.%s' % VERSION
 apilevel = '2.0'
 threadsafety = 1
@@ -43,8 +44,8 @@ def DEBUG_OUTPUT(*argv):
     if not DEBUG:
         return
     for s in argv:
-        print(s, end=' ', file=sys.stdout)
-    print(file=sys.stdout)
+        print(s, end=' ', file=sys.stderr)
+    print(file=sys.stderr)
 
 def HEX(data):
     import binascii
@@ -191,7 +192,7 @@ def _decode_column(data, oid, encoding):
     data = data.decode(encoding)
     if oid in (PG_TYPE_BOOL,):
         return data == 't'
-    elif oid in (PG_TYPE_INT2, PG_TYPE_INT4, PG_TYPE_INT8,):
+    elif oid in (PG_TYPE_INT2, PG_TYPE_INT4, PG_TYPE_INT8, PG_TYPE_OID,):
         return int(data)
     elif oid in (PG_TYPE_FLOAT4, PG_TYPE_FLOAT8):
         return float(data)
@@ -242,11 +243,16 @@ def _decode_column(data, oid, encoding):
         hex_str = data[2:]
         ia = [int(hex_str[i:i+2], 16) for i in range(0, len(hex_str), 2)]
         return b''.join([chr(c) for c in ia]) if PY2 else bytes(ia)
-    elif oid in (PG_TYPE_TEXT, PG_TYPE_VARCHAR, PG_TYPE_NAME, PG_TYPE_JSON):
+    elif oid in (PG_TYPE_TEXT, PG_TYPE_BPCHAR, PG_TYPE_VARCHAR, PG_TYPE_NAME, PG_TYPE_JSON):
         return data
+    elif oid in (PG_TYPE_UNKNOWN, ):
+        DEBUG_OUTPUT('UNKNOWN type')
+        return data
+    else:
+        if DEBUG:
+            raise ValueError(str(oid) + u":" + data)
 
     # other types return as string
-    DEBUG_OUTPUT('_decode_column()', data, oid)
     return data
 
 # ----------------------------------------------------------------------------
@@ -289,6 +295,14 @@ def escape_parameter(v):
         return "'" + ''.join(['\\%03o' % (c, ) for c in v]) + "'::bytea"
     elif t == bool:
         return u"'t'" if v else u"'f'"
+    elif t == time.struct_time:
+        return u'%04d-%02d-%02d %02d:%02d:%02d' % (
+            v.tm_year, v.tm_mon, v.tm_mday, v.tm_hour, v.tm_min, v.tm_sec)
+    elif t == datetime.timedelta:
+        if v.seconds:
+            return u"interval '%d second'" % (v.days * 86400 + v.seconds, )
+        else:
+            return u"interval '%d day'" % (v.days, )
     elif t == int or t == float or t == decimal.Decimal or (PY2 and t == long):
         return str(v)
     else:
@@ -444,7 +458,7 @@ class Connection(object):
         )
 
     def _process_messages(self, obj=None):
-        err = None
+        errobj = None
         while True:
             code = self._read(1)
             ln = _bytes_to_bint(self._read(4)) - 4
@@ -465,7 +479,7 @@ class Connection(object):
                     hash2 = hashlib.md5(hash1+salt).hexdigest().encode("ascii")
                     self._send_message(PG_F_PASSWORD_MESSAGE, b'md5'+hash2+'\x00')
                 else:
-                    err = InterfaceError("Authentication method %d not supported." % (auth_method,))
+                    errobj = InterfaceError("Authentication method %d not supported." % (auth_method,))
             elif code == PG_B_PARAMETER_STATUS:
                 k, v, _ = data.split(b'\x00')
                 k = k.decode('ascii')
@@ -544,9 +558,9 @@ class Connection(object):
                 errcode = err[1][1:].decode(self.encoding)
                 message = errcode + u':' + err[2][1:].decode(self.encoding)
                 if errcode[:2] == u'23':
-                    err = IntegrityError( message)
+                    errobj = IntegrityError( message)
                 else:
-                    err = DatabaseError(message)
+                    errobj = DatabaseError(message)
             elif code == PG_B_COPY_OUT_RESPONSE:
                 is_binary = data[0] == '\x01'
                 num_columns = _bytes_to_bint(data[1:3])
@@ -569,8 +583,8 @@ class Connection(object):
                 self._write(PG_COPY_DONE + b'\x00\x00\x00\x04' + PG_F_SYNC + b'\x00\x00\x00\x04')
             else:
                 DEBUG_OUTPUT("SKIP:", code, ln, HEX(data))
-        if err:
-            raise err
+        if errobj:
+            raise errobj
         return
 
     def __enter__(self):
@@ -618,17 +632,13 @@ class Connection(object):
     def _is_connect(self):
         return bool(self.sock)
 
-    def _close(self):
-        self.sock.close()
-        self.sock = None
 
     def cursor(self):
         return Cursor(self)
 
     def execute(self, query, obj):
         if not self.in_transaction:
-            self._send_message(PG_F_QUERY, b"BEGIN\x00")
-            self._process_messages(obj)
+            self.begin()
         DEBUG_OUTPUT('Connection::execute()', query)
         self._send_message(
             PG_F_QUERY,
@@ -636,8 +646,7 @@ class Connection(object):
         )
         self._process_messages(obj)
         if self.autocommit:
-            self._send_message(PG_F_QUERY, b"COMMIT\x00")
-            self._process_messages(obj)
+            self.commit()
 
     def set_autocommit(self, autocommit):
         self.autocommit = autocommit
@@ -649,9 +658,10 @@ class Connection(object):
 
     def commit(self):
         DEBUG_OUTPUT('COMMIT')
-        self._send_message(PG_F_QUERY, b"COMMIT\x00")
-        self._process_messages()
-        self.begin()
+        if self.sock:
+            self._send_message(PG_F_QUERY, b"COMMIT\x00")
+            self._process_messages()
+            self.begin()
 
     def rollback(self):
         DEBUG_OUTPUT('ROLLBACK')
@@ -668,7 +678,8 @@ class Connection(object):
         DEBUG_OUTPUT('Connection::close()')
         if self.sock:
             self._write(b''.join([PG_F_TERMINATE, b'\x00\x00\x00\x04']))
-            self._close()
+            self.sock.close()
+            self.sock = None
 
 def connect(host, user, password, database, port=5432, timeout=60, use_ssl=False):
     return Connection(user, password, database, host, port, timeout, use_ssl)
